#!/ur/bin/env python
# PYTHON_ARGCOMPLETE_OK
import os, sys

def main(args=None):
    import argparse, argcomplete, os, logging
    from collections import OrderedDict
    from plumbum import local
    from genairics import gscripts, typeMapping, logger, runWorkflow
    from genairics.jobs import QueuJob
    from genairics.RNAseq import RNAseq
    from genairics.ChIPseq import fastqcSample

    pipelines = OrderedDict((
        ('RNAseq',RNAseq),
        ('ChIPseq',fastqcSample)
    ))

    joblaunchers = OrderedDict((
        ('native', None),
        ('qsub', QueuJob)
    ))
    
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=r'''
      _______
      -------
         |||
      /+++++++:         __-_-_-_-_
  /#-'genairics'-,_     /;@;@,@'@+`
 /##@+;+++++++:@.+#\\/`,@@;@;@;@`@'\
 :#@:'###    #+';.#'\\@@'#####++;:@+
 |#@`:,:      #+#:+#+\\+#     #+@@++|
 ;@'.##         '+.:#_\\       #+@,`:
 :#@;.#         |+;,| ||        ##@`.|
 |@'..#         \\###-;|         #@:.:
 |#@;:#        '+\\::/\:         +#@::
 :@;,:+       #@@'\\/ ,\        #';@.|
 \#@'#'#     #'@',##;:@`        ##@,;;
  :##@:@:@;@;@;@.;/  \#+@;#` #'#+@`:;
  `\#.@;@;@;@:@.+/    :'@;@;@;@;@,:;
     \,,'::,,#::;      \'@:@'@;@'+'/

    GENeric AIRtight omICS pipeline.

    When the program is finished running, you can check the log file with "less -r plumbing/pipeline.log"
    from your project's result directory. Errors will also be printed to stdout.
    ''')

    parser.add_argument('--job-launcher', default = 'native', choices = joblaunchers.keys(),
                        help='choose where and how the job will run')
    parser.add_argument('--remote-host', default = '', help = 'submit job through ssh')
    parser.add_argument('--verbose', action = 'store_true', help = 'verbose output')

    # Pipeline subparsers
    subparsers = parser.add_subparsers(help='sub-command help')
    for pipeline in pipelines:
        subparser = subparsers.add_parser(
            pipeline,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='genairics {} -h'.format(pipeline)
        )
        subparser.set_defaults(function=pipelines[pipeline])
        for paran,param in pipelines[pipeline].get_params():
            if type(param._default) in typeMapping.values():
                subparser.add_argument('--'+paran, default=param._default, type=typeMapping[type(param)],
                                       help=param.description)
            else: subparser.add_argument(paran, type=typeMapping[type(param)], help=param.description)

    # Console option
    def startConsole():
        import IPython
        import genairics, genairics.datasources, genairics.resources
        from genairics.resources import InstallDependencies
        IPython.embed()
        exit()
        
    subparser = subparsers.add_parser(
        'console',
        help='Start console where tasks can be started not available in the commandline interface'
    )
    subparser.set_defaults(function=startConsole)
    
    if args is None:
        # if arguments are set in environment, they are used as the argument default values
        # this allows seemless integration with PBS jobs
        if 'GENAIRICS_ENV_ARGS' in os.environ:
            #Retrieve arguments from qsub job environment
            args = [os.environ['GENAIRICS_ENV_ARGS']]
            positionals = []
            optionals = []
            for paran,param in pipelines[args[0]].get_params():
                if paran in os.environ:
                    if type(param._default) in typeMapping.values():
                        optionals += ['--'+paran, os.environ[paran]]
                    else: positionals.append(os.environ[paran])
            logger.warning(
                'Pipeline %s arguments were retrieved from environment: positional %s, optional %s',
                args[0], positionals, optionals
            )
            args+= optionals + positionals
            args = parser.parse_args(args)
        else:
            #Script started directly
            argcomplete.autocomplete(parser)
            args = parser.parse_args()
        #Make dict out of args namespace for passing to pipeline
        args = vars(args)

    joblauncher = joblaunchers[args.pop('job_launcher')]
    remotehost = args.pop('remote_host')
    verbose = args.pop('verbose')
    workflow = args.pop('function')(**args)

    if verbose:
        logger.setLevel(logging.DEBUG)
    else: logger.setLevel(logging.INFO)
    
    if joblauncher:
        joblauncher(job=workflow,remote=remotehost).run()
    else: runWorkflow(workflow)

# If only program name is entered generate GUI
if len(sys.argv) == 1:
    from gooey import Gooey
    main = Gooey(
        advanced=True,
        program_description='genairics: generic airtight omics pipelines'
    )(main)
    # Disable stdout buffering so gooey has smooth output
    nonbuffered_stdout = os.fdopen(sys.stdout.fileno(), 'w', 1)
    sys.stdout = nonbuffered_stdout

# Run main program logics when script called directly
if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Tasks here prepare resources that are required for some
pipeline tasks, and are generally available from the
resources directory as specified by $GAX_RESOURCES
"""
from datetime import datetime, timedelta
import luigi, os, sys, tempfile, pathlib, glob
from luigi.util import inherits
from plumbum import local, colors
import logging

from genairics import gscripts

resourcedir = os.environ.get('GAX_RESOURCES',os.path.expanduser('~/resources'))
logresources = logging.Logger('genairics.resources')

#Set cksum program according to platform
if sys.platform == 'linux':
    cksum = local['sum']
elif sys.platform == 'darwin':
    cksum = local['cksum']['-o',1]
else: raise Exception('{} not supported. Run from docker container.'.format(sys.platform))

def requestFiles(urlDir,fileregex,outdir):
    """
    Download a set of files that match fileregex from urlDir
    checksums are expected to be generated by the "sum" command,
    as is the case for the ftp.ensembl.org repository
    """
    import requests, re
    link = re.compile(r'href="(.+?)"')
    if fileregex: fileregex = re.compile(fileregex)
    if not urlDir.endswith('/'): urlDir+='/'
    # Download checksums if present
    checksums = requests.get(urlDir+'CHECKSUMS')
    if checksums:
        from io import StringIO
        import pandas as pd
        checksums = pd.read_table(StringIO(checksums.text),sep='\s+',
                                  names=['checksum', 'octets', 'name'],index_col='name')
    else: checksums = None
    # Retrieve index
    r = requests.get(urlDir)
    for l in link.finditer(r.text):
        l = l.groups()[0]
        if not fileregex or fileregex.match(l):
            file_request = requests.get(urlDir+l, stream=True)
            with open(os.path.join(outdir,l),'wb') as outfile:
                for chunk in file_request.iter_content(chunk_size=512):
                    outfile.write(chunk)
            if checksums is not None:
                csum = int(cksum(os.path.join(outdir,l)).split()[0])
                if csum != checksums.checksum[l]:
                    logresources.warning('%s checksum did not match url location %s',csum,urlDir)

# Tasks
class InstallDependencies(luigi.Task):
    """
    Installs the genairics dependency programs
    """
    def run(self):
        local[gscripts % 'genairics_dependencies.sh']
    
class RetrieveGenome(luigi.Task):
    """
    Prepares the genome
    """
    genome = luigi.Parameter(default='homo_sapiens',description="genome species name")
    release = luigi.IntParameter(default=91,description="ensembl release number of genome")

    def output(self):
        return luigi.LocalTarget(
            os.path.join(resourcedir,'ensembl',self.genome,'release-{}'.format(self.release))
        )

    def run(self):
        #Make temp dir for data
        local['mkdir']('-p',self.output().path+'_retrieving/dna')
        local['mkdir']('-p',self.output().path+'_retrieving/annotation')
        requestFiles(
            'http://ftp.ensembl.org/pub/release-{release}/fasta/{species}/dna/'.format(
                species=self.genome, release=self.release),
            r'.+.dna.chromosome.+.fa.gz',
            self.output().path+'_retrieving/dna'
        )
        requestFiles(
            'http://ftp.ensembl.org/pub/release-{release}/gtf/{species}/'.format(
                species=self.genome, release=self.release),
            r'.+.{release}.gtf.gz'.format(release=self.release),
            self.output().path+'_retrieving/annotation'
        )
        #Unzip all files
        local['gunzip'](*glob.glob(self.output().path+'_retrieving/*/*.gz'))
        #Rename temp dir to final/expected output dir
        os.rename(self.output().path+'_retrieving',self.output().path)

@inherits(RetrieveGenome)
class STARandRSEMindex(luigi.Task):
    """
    Index that can be used by both STAR aligner and RSEM counter for transcriptomics
    """
    threads = luigi.IntParameter(default=1, description='STAR threads to use to build mapping index')

    def requires(self):
        return self.clone_parent()
    
    def output(self):
        return (
            luigi.LocalTarget(
                os.path.join(resourcedir,'ensembl/{species}/release-{release}/transcriptome_index'.format(
                    species=self.genome,release=self.release))
            ),
            luigi.LocalTarget(
                os.path.join(resourcedir,'ensembl/{species}/release-{release}/transcriptome_index/build_completed'.format(
                    species=self.genome,release=self.release))
            )
        )
    
    def run(self):
        genomeDir = self.input().path
        os.mkdir(self.output()[0].path)
        stdout = local[gscripts % 'buildRSEMindex.sh'](
            *glob.glob(os.path.join(genomeDir,'annotation')+'/*.gtf'),
            self.threads,
            ','.join(glob.glob(os.path.join(genomeDir,'dna')+'/*.fa*')),
            os.path.join(self.output()[0].path, self.genome)
        )
        logresources.info(stdout)
        pathlib.Path(self.output()[1].path).touch()
